/* Generate a program to test NEON instructions.
   Copyright (c) 2006 CodeSourcery
   Written by Paul Brook.
   This code is licenced under the GNU GPL v2 or later.  */

#include <stdio.h>
#include <stdbool.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>

//#define DEBUG

#ifdef DEBUG
#define DPRINTF(args...) printf(args)
#else
#define DPRINTF(args...) do { } while(0)
#endif

int lineno;

void
die(const char *msg, ...)
{
    va_list va;
    va_start(va, msg);
    fprintf(stderr, "Error (%d):", lineno);
    vfprintf(stderr, msg, va);
    fprintf(stderr, "\n");
    va_end(va);
    exit(1);
}

typedef struct
{
  char name[128];
  bool elemental;
  bool pairwise;
  bool wide;
  bool narrow;
  bool lng;
  bool both;
  bool zero;
  bool accumulate;
  int inputs;
  int outputs;
} insn_state;

#define END "\\n\"\n\""
static void
load_reg(FILE *outfile, int pos, int reg, bool quad)
{
  if (quad)
    fprintf (outfile, "vldmia %%%d, {q%d}"END, pos, reg);
  else
    fprintf (outfile, "vldr d%d, [%%%d]"END, reg, pos);
}

static void
save_reg(FILE *outfile, int pos, int reg, bool quad)
{
  if (quad)
    fprintf (outfile, "vstmia %%%d, {q%d}"END, pos, reg);
  else
    fprintf (outfile, "vstr d%d, [%%%d]"END, reg, pos);
}

void
write_asm(FILE *outfile, insn_state *s, const char *typename, int quad,
	  const int *argno)
{
  int n;
  char c, d;
  const char *asmname;

  asmname = s->name;
  if (s->zero)
    asmname += 2;

  fprintf (outfile, "asm volatile(\"");
  /* Source width.  */
  c = (quad || s->narrow) ? 'q' : 'd';
  /* Destination width.  */
  d = (quad || s->lng) ? 'q' : 'd';
  if (s->accumulate)
    {
      load_reg (outfile, 4, argno[0], d == 'q');
      load_reg (outfile, 2, argno[1], c == 'q');
      if (s->inputs == 2)
	{
	  load_reg (outfile, 3, argno[2], c == 'q');
	  fprintf (outfile, "v%s.%s %c%d, %c%d, %c%d"END, asmname, typename,
		   d, argno[0], c, argno[1], c, argno[2]);
	}
      else
	{
	  fprintf (outfile, "v%s.%s %c%d, %c%d"END, asmname, typename,
		   d, argno[0], c, argno[1]);
	}
    }
  else if (s->inputs == 1)
    {
      load_reg (outfile, 2, argno[1], c == 'q');
      fprintf (outfile, "v%s.%s %c%d, %c%d%s"END, asmname, typename,
	       d, argno[0], c, argno[1], s->zero ? ", #0" : "");
    }
  else if (s->outputs == 2)
    {
      load_reg (outfile, 2, argno[0], quad);
      load_reg (outfile, 3, argno[1], quad);
      fprintf (outfile, "v%s.%s %c%d, %c%d"END, asmname, typename,
	       d, argno[0], c, argno[1]);
    }
  else if (s->inputs == 2)
    {
      load_reg (outfile, 2, argno[1], c == 'q');
      load_reg (outfile, 3, argno[2], c == 'q');
      fprintf (outfile, "v%s.%s %c%d, %c%d, %c%d"END, asmname, typename,
	       d, argno[0], c, argno[1], c, argno[2]);
    }
  else
    abort();
  save_reg (outfile, 0, argno[0], d == 'q');
  if (s->outputs == 2)
    save_reg (outfile, 1, argno[1], quad);
  fprintf (outfile, "\":: \"r\"(res0), \"r\"(res1),\n"
	  "\"r\"(arg0), \"r\"(arg1), \"r\"(arg2) : \"memory\");\n");
  fprintf(outfile, "fail = 0;\n");
  fprintf (outfile, "if(res0[0] != em0[0]) fail=1;\n");
  if (quad || s->lng)
    fprintf (outfile, "if(res0[1] != em0[1]) fail=1;\n");
  if (s->outputs == 2)
    {
      fprintf (outfile, "if(res1[0] != em1[0]) fail=1;\n");
      if (quad || s->lng)
	fprintf (outfile, "if(res1[1] != em1[1]) fail=1;\n");
    }
  fprintf (outfile, "if (fail) { printf (\"FAIL: %s_%s:\\n\");\n",
	   s->name, typename);
  if (typename[0] == 'f')
    c = d = 'f';
  fprintf (outfile, "dump_arg%c(em0, \"em0 \");\n", d);
  if (s->outputs == 2)
    fprintf (outfile, "dump_arg%c(em1, \"em1 \");\n", d);
  fprintf (outfile, "dump_arg%c(res0, \"out0\");\n", d);
  if (s->outputs == 2)
    fprintf (outfile, "dump_arg%c(res1, \"out1\");\n", d);
  fprintf (outfile, "dump_arg%c(arg0, \"arg0\");\n", c);
  if (s->inputs > 1)
    fprintf (outfile, "dump_arg%c(arg1, \"arg1\");\n", c);
  if (s->accumulate)
    fprintf (outfile, "dump_arg%c(arg2, \"arg2\");\n", d);
  fprintf (outfile, "nfail++;\n}\n");
  fprintf (outfile, "else printf (\"PASS: %s_%s\\n\");\n",
	   s->name, typename);
}

void
write_em(FILE *outfile, insn_state *s, const char *typename, bool quad)
{
  if (s->accumulate)
    {
      fprintf (outfile, "arg2[0] = get_value_%s();\n", typename);
      if (quad || s->lng)
	fprintf (outfile, "arg2[1] = get_value_%s();\n", typename);
    }
  fprintf (outfile, "arg0[0] = get_value_%s();\n", typename);
  if (quad)
    fprintf (outfile, "arg0[1] = get_value_%s();\n", typename);
  if (s->inputs > 1)
    {
      fprintf (outfile, "arg1[0] = get_value_%s();\n", typename);
      if (quad)
	fprintf (outfile, "arg1[1] = get_value_%s();\n", typename);
    }
  fprintf (outfile, "emulate_%s_%s(em0, ", s->name, typename);
  if (s->outputs == 2)
    fprintf (outfile, "em1, ");
  if (s->accumulate)
    fprintf (outfile, "arg2, ");
  if (s->inputs == 1)
    fprintf (outfile, "arg0[0]);\n");
  else if (quad && s->pairwise)
    fprintf (outfile, "arg0[0], arg0[1]);\n");
  else
    fprintf (outfile, "arg0[0], arg1[0]);\n");
  if (quad)
    {
      fprintf (outfile, "emulate_%s_%s(em0 + 1, ", s->name, typename);
      if (s->outputs == 2)
	fprintf (outfile, "em1 + 1, ");
      if (s->accumulate)
	fprintf (outfile, "arg2 + 1, ");
      if (s->inputs == 1)
	fprintf (outfile, "arg0[1]);\n");
      else if (quad && s->pairwise)
	fprintf (outfile, "arg1[0], arg1[1]);\n");
      else
	fprintf (outfile, "arg0[1], arg1[1]);\n");
    }
}

static const int argno_unop[2][3] = {
    {0, 0}, {0, 1}
};
static const int argno_swap[1][3] = {
    {0, 1}
};
/* ??? This does not test overlapping arguments.  */
static const int argno_arith[3][3] = {
    {0, 0, 1}, {0, 1, 0}, {0, 1, 2}
};
static const int argno_tri[1][3] = {
    {0, 2, 3}
};
static const int argno_long[17][3] = {
    {0, 0, 1}, {0, 0, 2}, {0, 0, 3}, {0, 0, 4},
    {0, 1, 0}, {0, 1, 3}, {0, 1, 3}, {0, 1, 4},
    {0, 2, 0}, {0, 2, 1}, {0, 2, 3}, {0, 2, 4},
    {0, 3, 0}, {0, 3, 1}, {0, 3, 2}, {0, 3, 4},
    {0, 4, 5}
};

void
write_test(FILE *outfile, insn_state *s, const char *typename)
{
  int n;
  bool quad = false;

  fprintf (outfile, "static void\ndo_test_%s_%s(void)\n{\n", s->name, typename);
  fprintf (outfile, "u64 arg0[2], arg1[2], arg2[2];\n");
  fprintf (outfile, "u64 res0[2], res1[2];\n");
  fprintf (outfile, "u64 em0[2], em1[2];\n");
  fprintf (outfile, "int fail;\n");
again:
  write_em(outfile, s, typename, quad);
  if (s->accumulate)
    {
      for (n = 0; n < 1; n++)
	write_asm (outfile, s, typename, quad, argno_tri[n]);
    }
  else if (s->lng)
    {
      for (n = 0; n < 17; n++)
	write_asm (outfile, s, typename, quad, argno_long[n]);
    }
  else if (s->outputs == 2)
    {
      write_asm (outfile, s, typename, quad, argno_swap[0]);
    }
  else if (s->inputs == 1)
    {
      for (n = 0; n < 2; n++)
	write_asm (outfile, s, typename, quad, argno_unop[n]);
    }
  else
    {
      for (n = 0; n < 3; n++)
	write_asm (outfile, s, typename, quad, argno_arith[n]);
    }
  if (s->both && !quad)
    {
      quad = true;
      goto again;
    }
  fprintf (outfile, "}\n\n");
}

typedef struct name_list
{
  struct name_list *next;
  char name[];
} name_list;

static name_list *names;
static name_list **name_tail = &names;

static void
add_test(insn_state *s, const char *typename)
{
  name_list *p;
  int len;
  len = strlen(s->name) + strlen(typename) + 2;
  p = malloc(len + sizeof(name_list));
  sprintf(p->name, "%s_%s", s->name, typename);
  *name_tail = p;
  p->next = NULL;
  name_tail = &p->next;
}

static void
write_main(FILE *outfile)
{
  name_list *p;

  fprintf (outfile, "int main(int argc, char ** argv)\n{\n");
  fprintf (outfile, "helper_init(argc, argv);\n");
  for (p = names; p; p = p->next)
    {
      fprintf (outfile, "do_test_%s();\n", p->name);
    }
  fprintf (outfile, "printf(\"%%d failures\\n\", nfail);\n");
  fprintf (outfile, "return nfail != 0;\n}\n");
}

static void
do_test_1(FILE *outfile, insn_state *s, const char *typename, const char *insn)
{
  int typesize;
  int nelem;
  const char *argtype;
  int i;
  char desttype[4];

  DPRINTF("do_test_1: %s_%s %s\n", s->name, typename, insn);
  typesize = atoi (typename + 1);
  nelem = 64 / typesize;
  if (s->lng)
    sprintf(desttype, "%c%d", typename[0], typesize * 2);
  else
    strcpy(desttype, typename);
  fprintf (outfile, "static void\nemulate_%s_%s(", s->name, typename);
  if (s->pairwise)
    {
      if (s->lng)
	die("pairwise long\n");
      fprintf (outfile, "u64 *pdest, ");
      fprintf (outfile, "u64 vsrc1, u64 vsrc2)\n{\n");
      fprintf (outfile, "u64 vdest;\n");
      fprintf (outfile, "%s dest, src1, src2;\n", typename);
      fprintf (outfile, "int n;\n");
      fprintf (outfile, "for (n = 0; n < %d; n++) {\n", nelem / 2);
      fprintf (outfile, "src1 = EXTRACT(%s, vsrc1, n * 2);\n", typename);
      fprintf (outfile, "src2 = EXTRACT(%s, vsrc1, n * 2 + 1);\n", typename);
      fprintf (outfile, "{%s;}\n", insn);
      fprintf (outfile, "INSERT(%s, vdest, dest, n);\n", typename);
      fprintf (outfile, "}\n");
      fprintf (outfile, "for (n = 0; n < %d; n++) {\n", nelem / 2);
      fprintf (outfile, "src1 = EXTRACT(%s, vsrc2, n * 2);\n", typename);
      fprintf (outfile, "src2 = EXTRACT(%s, vsrc2, n * 2 + 1);\n", typename);
      fprintf (outfile, "{%s;}\n", insn);
      fprintf (outfile, "INSERT(%s, vdest, dest, n + %d);\n",
	       typename, nelem / 2);
      fprintf (outfile, "}\n");
      fprintf (outfile, "*pdest = vdest;\n");
    } 
  else
    {
      if (s->outputs == 2)
	fprintf (outfile, "u64 *pdest1, u64 *pdest2, ");
      else
	fprintf (outfile, "u64 *pdest, ");
      if (s->accumulate)
	fprintf (outfile, "u64 *psrcdest, ");
      if (s->inputs == 1)
	fprintf (outfile, "u64 %ssrc", s->elemental ? "v" : "");
      else
	fprintf (outfile, "u64 %ssrc1, u64 %ssrc2",
		 s->elemental ? "v" : "", s->elemental ? "v" : "");
      fprintf (outfile, ")\n{\n");

      if (s->elemental)
	{
	  if (s->outputs == 2)
	    fprintf (outfile, "u64 vdest1, dest2;\n");
	  else
	    fprintf (outfile, "u64 vdest;\n");
	  if (s->accumulate)
	    fprintf (outfile, "u64 vsrcdest = *(psrcdest++);\n");
	  fprintf (outfile, "int n, dn;\n");
	  fprintf (outfile, "for (n = 0, dn = 0; n < %d; n++) {\n", nelem);
	  if (s->inputs == 1)
	    {
	      fprintf (outfile, "%s src;\n", typename);
	      fprintf (outfile, "%s dest;\n", desttype);
	      fprintf (outfile, "src = EXTRACT(%s, vsrc, n);\n", typename);
	    }
	  else
	    {
	      fprintf (outfile, "%s src1, src2;\n", typename);
	      if (s->outputs == 2)
		fprintf (outfile, "%s dest1, dest2;\n", desttype);
	      else
		fprintf (outfile, "%s dest;\n", desttype);
	      fprintf (outfile, "src1 = EXTRACT(%s, vsrc1, n);\n", typename);
	      fprintf (outfile, "src2 = EXTRACT(%s, vsrc2, n);\n", typename);
	    }
	  if (s->accumulate)
	    fprintf (outfile, "dest = EXTRACT(%s, vsrcdest, dn);\n", desttype);
	  fprintf (outfile, "{%s;}\n", insn);
	  if (s->outputs == 1)
	    fprintf (outfile, "INSERT(%s, vdest, dest, dn);\n", desttype);
	  else
	    {
	      fprintf (outfile, "INSERT(%s, vdest1, dest1, dn);\n", desttype);
	      fprintf (outfile, "INSERT(%s, vdest2, dest2, dn);\n", desttype);
	    }
	  if (s->lng)
	    {
	      fprintf (outfile, "if (n == %d) {\n", (nelem / 2) - 1);
	      if (s->outputs == 2)
		fprintf (outfile, "*(pdest1++) = vdest1; *(pdest2++) = vdest2;\n");
	      else
		fprintf (outfile, "*(pdest++) = vdest;\n");
	      if (s->accumulate)
		fprintf (outfile, "vsrcdest = *(psrcdest++);\n");
	      fprintf (outfile, "dn = 0;\n");
	      fprintf (outfile, "} else\n");
	    }
	  fprintf (outfile, "dn++;\n");
	  fprintf (outfile, "}\n");
	  if (s->outputs == 2)
	    fprintf (outfile, "*pdest1 = vdest1; *pdest2 = vdest2;\n");
	  else
	    fprintf (outfile, "*pdest = vdest;\n");
	}
      else
	{
	  if (s->lng)
	    die("Non-elemental long\n");
	  if (s->outputs == 2)
	    fprintf (outfile, "u64 dest1, dest2;\n");
	  else
	    fprintf (outfile, "u64 dest;\n");
	  if (s->accumulate)
	    fprintf (outfile, "dest = *psrcdest;\n");
	  fprintf (outfile, "{%s;}\n", insn);
	  if (s->outputs == 2)
	    fprintf (outfile, "*pdest1 = dest1, *pdest2 = dest2;\n");
	  else
	    fprintf (outfile, "*pdest = dest;\n");
	}
    }
  fprintf (outfile, "}\n\n");
  write_test(outfile, s, typename);
  add_test(s, typename);
}

void
do_test(FILE *outfile, insn_state *s, const char *typespec, const char *insn)
{
  char type[16];
  char c;
  int n;

  /* TODO: implement wide and narrow ops.  */
  if (s->wide || s->narrow)
    return;
dot:
  if (*typespec != '.')
    die("Expected '.', got '%s'", typespec);
  typespec++;
again:
  c = *(typespec++);
  type[0] = c;
  n = 1;

  while (isdigit(*typespec))
    type[n++] = *(typespec++);

  if (n == 1)
    {
      if (c == 'f')
	{
	  strcpy(type + 1, "32");
	  do_test_1(outfile, s, type, insn);
	}
      else if (c == 'u' || c == 's')
	{
	  strcpy(type + 1, "8");
	  do_test_1(outfile, s, type, insn);
	  strcpy(type + 1, "16");
	  do_test_1(outfile, s, type, insn);
	  strcpy(type + 1, "32");
	  do_test_1(outfile, s, type, insn);
	}
      else if (c == 'p')
	{
	  strcpy(type + 1, "8");
	  do_test_1(outfile, s, type, insn);
	}
      else
	die ("Bad type '%c'", c);
    }
  else
    {
      type[n] = 0;
      do_test_1(outfile, s, type, insn);
    }
  c = *(typespec);
  if (!c)
    return;
  if (c != ' ')
    goto again;
  typespec++;
  goto dot;
}

static void
write_file_header(FILE *outfile)
{
  fprintf(outfile, "#include \"neon-helper.c\"\n");
  /* Nasty hack for compilers that don't support NEON.  When using a
     neon/fpa aware compiler we should add clobbers to the asms.  */
  fprintf(outfile, "asm (\".fpu neon\");\n");
}

int
main(int argc, char **argv)
{
    FILE *f;
    FILE *dotestfile;
    char line[1024];
    char *p;
    int len;
    insn_state *s;

    if (argc != 3)
      {
	printf ("Usage: gen-tests [infile] [outfile]");
	exit(1);
      }
    s = malloc (sizeof(*s));
    f = fopen (argv[1], "rt");

    dotestfile = fopen(argv[2], "wt");

    lineno = 0;
    write_file_header(dotestfile);
    while (fgets (line, 1024, f))
      {
	lineno++;
	if (line[0] == '#')
	  continue;
	len = strlen(line) - 1;
	/* Remove trailing \n */
	line[len] = 0;
	switch (line[0]) {
	  case 0:
	    /* No-op */
	    break;
	  case '*':
	      memset(s, 0, sizeof(*s));
	      strcpy(s->name, line + 1);
	      s->inputs = 2;
	      s->outputs = 1;
	      break;
	  case '+':
	      for (p = line + 1; *p; p++)
		{
		  switch (*p) {
		    case 'e': s->elemental = true; break;
		    case 'p': s->pairwise = true; break;
		    case 'l': s->lng = true; break;
		    case 'b': s->both = true; break;
		    case 'd': break;
		    case 'n': s->narrow = true; break;
		    case 'w': s->wide = true; break;
		    case '1': s->inputs = 1; break;
		    case 'a': s->accumulate = true; break;
		    case 'x': s->outputs = 2; break;
		    case 'z': s->zero = true; s->inputs = 1; break;
		    case ',': break;
		    default: die ("Unrecognised option '%c'", *p);
		  }
		}
	      break;
	  case '.':
	      p = strchr(line, ':');
	      *(p++) = 0;
	      do_test(dotestfile, s, line, p);
	      break;
	}
      }
    write_main(dotestfile);
    fclose(f);
    fclose(dotestfile);
}
