/* Helper routines for NEON tests.
   Copyright (c) 2006 CodeSourcery
   Written by Paul Brook.
   This code is licenced under the GNU GPL v2 or later.  */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "test-neon.h"

int nfail = 0;

#define EXTRACT(type, src, n) extract_##type(src, n)
#define INSERT(type, dest, val, n) dest = insert_##type(dest, val, n)

#define DO_EXTRACT(type) \
static inline type \
extract_##type(u64 vec, int n) \
{ \
  return vec >> (n * sizeof(type) * 8); \
}
DO_EXTRACT(s8)
DO_EXTRACT(u8)
DO_EXTRACT(s16)
DO_EXTRACT(u16)
DO_EXTRACT(s32)
DO_EXTRACT(u32)
#undef DO_EXTRACT
#define extract_u64(vec, n) (vec)
#define extract_s64(vec, n) (s64)(vec)
#define extract_p8 extract_u8

static inline f32
extract_f32(u64 vec, int n)
{
  union {
    f32 f[2];
    u64 i;
  } u;
  u.i = vec;
  return u.f[n];
}

#define DO_INSERT(bits) \
static inline u64 \
insert_u##bits(u64 vec, u##bits val, int n) \
{ \
  u64 tmp; \
  u64 mask; \
  mask = ((u64)1u << bits) - 1; \
  mask <<= (n * bits); \
  tmp = vec & ~mask; \
  tmp |= ((u64)val << (n * bits)); \
  return tmp; \
}
DO_INSERT(8)
DO_INSERT(16)
DO_INSERT(32)
#undef DO_INSERT
#define insert_s8 insert_u8
#define insert_s16 insert_u16
#define insert_s32 insert_u32
#define insert_u64(vec, val, n) (u64)(val)
#define insert_s64(vec, val, n) (s64)(val)
#define insert_p8 insert_u8

static inline u64
insert_f32(u64 vec, f32 val, int n)
{
  union {
    f32 f[2];
    u64 i;
  } u;
  u.i = vec;
  u.f[n] = val;
  return u.i;
}

#define DEF_ABD(type) \
  type abd_##type(type a, type b) \
  {return (a > b) ? a - b : b - a;}
DEF_ABD(u16)
DEF_ABD(s16)
DEF_ABD(u32)
DEF_ABD(s32)
DEF_ABD(u64)
DEF_ABD(s64)
#undef DEF_ABD

static inline p8
mul_p8(p8 src1, p8 src2)
{
  p8 res;

  res = 0;
  while(src1)
    {
      if (src1 & 1)
	res ^= src2;
      src1 >>= 1;
      src2 <<= 1;
    }
  return res;
}

static inline f32
fcmp(int val)
{
  union {
    u32 i;
    f32 f;
  } u;
  u.i = val ? -1 : 0;
  return u.f;
}

u32 do_cls(u32 val, int bits)
{
    u32 mask;
    u32 sign;
    int n;

    sign = 1 << (bits - 1);
    mask = sign >> 1;
    sign &= val;
    sign >>= 1;
    n = 0;
    while (mask && (mask & val) == sign) {
	mask >>= 1;
	sign >>= 1;
	n++;
    }
    return n;
}

u32 do_clz(u32 val, int bits)
{
    u32 mask;
    int n;

    mask = 1 << (bits - 1);
    n = 0;
    while (mask && (val & mask) == 0) {
	mask >>= 1;
	n++;
    }
    return n;
}

u32 do_bcnt8(u8 val)
{
    int n;
    int count;

    count = 0;
    for (n = 0; n < 8; n++) {
	if (val & (1 << n))
	    count++;
    }
    return count;
}

#define DO_PADD(tn, tw, nelem) \
u64 padd_##tn(uint64_t src) \
{ \
  tw tmp1; \
  tw tmp2; \
  u64 dest; \
  int n; \
  for (n = 0; n < nelem; n++) \
    { \
      tmp1 = EXTRACT(tn, src, n * 2); \
      tmp2 = EXTRACT(tn, src, n * 2 + 1); \
      tmp1 += tmp2; \
      INSERT(tw, dest, tmp1, n); \
    } \
  return dest; \
}
DO_PADD(u8, u16, 4)
DO_PADD(s8, u16, 4)
DO_PADD(u16, u32, 4)
DO_PADD(s16, u32, 4)
DO_PADD(u32, u64, 4)
DO_PADD(s32, u64, 4)
#undef DO_PADD

static inline u64
paddl_u8(u64 src)
{
  u64 tmp;
  tmp = EXTRACT(u8, src, 7);
  tmp += EXTRACT(u8, src, 6);
  tmp <<= 16;
  tmp += EXTRACT(u8, src, 5);
  tmp += EXTRACT(u8, src, 4);
  tmp <<= 16;
  tmp += EXTRACT(u8, src, 3);
  tmp += EXTRACT(u8, src, 2);
  tmp <<= 16;
  tmp += EXTRACT(u8, src, 1);
  tmp += EXTRACT(u8, src, 0);
  return tmp;
}

static inline u64
paddl_u16(u64 src)
{
  u64 tmp;
  tmp = EXTRACT(u16, src, 3);
  tmp += EXTRACT(u16, src, 2);
  tmp <<= 32;
  tmp += EXTRACT(u16, src, 1);
  tmp += EXTRACT(u16, src, 0);
  return tmp;
}

static inline u64
paddl_u32(u64 src)
{
  u64 tmp;
  tmp = EXTRACT(u32, src, 1);
  tmp += EXTRACT(u32, src, 0);
  return tmp;
}

static inline u64
paddl_s8(u64 src)
{
  u64 res;
  u16 tmp;
  tmp = EXTRACT(s8, src, 7);
  tmp += EXTRACT(s8, src, 6);
  res = tmp;
  tmp = EXTRACT(s8, src, 5);
  tmp += EXTRACT(s8, src, 4);
  res = (res << 16) | tmp;
  tmp = EXTRACT(s8, src, 3);
  tmp += EXTRACT(s8, src, 2);
  res = (res << 16) | tmp;
  tmp = EXTRACT(s8, src, 1);
  tmp += EXTRACT(s8, src, 0);
  res = (res << 16) | tmp;
  return res;
}

static inline u64
paddl_s16(u64 src)
{
  u32 low, high;
  low = EXTRACT(s16, src, 1);
  low += EXTRACT(s16, src, 0);
  high = EXTRACT(s16, src, 3);
  high += EXTRACT(s16, src, 2);
  return low | ((u64)high << 32);
}

static inline u64
paddl_s32(u64 src)
{
  s64 tmp;
  tmp = EXTRACT(s32, src, 1);
  tmp += EXTRACT(s32, src, 0);
  return tmp;
}

static inline u64
add_u16(u64 a, u64 b)
{
  u64 res;
  u16 tmp;
  int n;
  for (n = 0; n < 4; n++)
    {
      tmp = EXTRACT(u16, a, n) + EXTRACT(u16, b, n);
      INSERT(u16, res, tmp, n);
    }
  return res;
}

static inline u64
add_u32(u64 a, u64 b)
{
  u64 res;
  u32 tmp;
  tmp = EXTRACT(u32, a, 0) + EXTRACT(u32, b, 0);
  res = EXTRACT(u32, a, 1) + EXTRACT(u32, b, 1);
  return (res << 32) | tmp;
}

#define get_value_u8 get_value
#define get_value_s8 get_value
#define get_value_p8 get_value
#define get_value_u16 get_value
#define get_value_s16 get_value
#define get_value_u32 get_value
#define get_value_s32 get_value
#define get_value_u64 get_value
#define get_value_s64 get_value

static void
helper_init(int argc, char ** argv)
{
  srand(0);
#if 1
  /* If running on bare metal  we may need to enable the coprocessor.  */
  asm volatile (
    /* Skip this code if in user mode.  */
    "mrs r0, cpsr\n"
    "and r0, r0, #0x1f\n"
    "cmp r0, #0x10\n"
    "beq 1f\n"
    "MRC p15, 0, r1, c1, c0, 2\n" /* r1 = Access Control Register */
    "ORR r1, r1, #(0xf << 20)\n"  /* enable full access for p10,11 */
    "MCR p15, 0, r1, c1, c0, 2\n" /* Access Control Register = r1 */
    "MOV r1, #0\n"
    "MCR p15, 0, r1, c7, c5, 4\n" /* flush prefetch buffer because of FMXR */
    /* Enable VFP itself */
    "MOV     r0,#0x40000000\n"
    "FMXR    FPEXC, r0\n"	  /* FPEXC = r0 */
    "1:"
    :::"r0", "r1");
#endif
  printf ("Running tests\n");
}

static u64
get_value(void)
{
  u64 tmp;
  /* ??? Maybe make this smarted by choosing values that are known
     corner cases. */
  tmp = (u16)rand();
  tmp = (tmp << 16) | (u16)rand();
  tmp = (tmp << 16) | (u16)rand();
  tmp = (tmp << 16) | (u16)rand();
  return tmp;
}

static u64
get_value_f32(void)
{
  union {
      u64 i;
      f32 f[2];
  } u;
  u.f[0] = (f32)(s16)rand();
  u.f[1] = (f32)(s16)rand();
  return u.i;
}

static void
dump_argf(u64 *v, const char * s)
{
  union {
      u64 i;
      f32 f[2];
  } u;
  u.i = v[0];
  printf ("%s=%08x%08x %g %g\n", s, (int)(v[0] >> 32), (int)v[0],
	  u.f[1], u.f[0]);
}

static void
dump_argd(u64 *v, const char * s)
{
  printf ("%s=%08x%08x\n", s, (int)(v[0] >> 32), (int)v[0]);
}

static void
dump_argq(u64 *v, const char * s)
{
  printf ("%s=%08x%08x%08x%08x\n", s, (int)(v[1] >> 32), (int)v[1],
	  (int)(v[0] >> 32), (int)v[0]);
}


#define do_sat8(arg0, op, arg1) ({ \
    s16 sat_tmp; \
    sat_tmp = (s16)(s8)(arg0) op (s16)(s8)(arg1); \
    if (sat_tmp != (s8)sat_tmp) \
      sat_tmp = (sat_tmp >> 15) ^ 0x7f; \
    (s8)sat_tmp; \
    })

#define do_sat16(arg0, op, arg1) ({ \
    s32 sat_tmp; \
    sat_tmp = (s32)(s16)(arg0) op (s32)(s16)(arg1); \
    if (sat_tmp != (s16)sat_tmp) \
      sat_tmp = (sat_tmp >> 31) ^ 0x7fff; \
    (s16)sat_tmp; \
    })

#define do_sat32(op0, op, op1) ({ \
    s64 sat_tmp; \
    sat_tmp = (s64)(s32)(op0) op (s64)(s32)(op1); \
    if (sat_tmp != (s32)sat_tmp) \
      sat_tmp = (sat_tmp >> 63) ^ 0x7fffffff; \
    (s32)sat_tmp; \
    })

#define do_usat8(op0, op, op1) ({ \
    s16 sat_tmp; \
    sat_tmp = (s16)(u8)(op0) op (s16)(u8)(op1); \
    if (sat_tmp != (u8)sat_tmp) \
      sat_tmp = (sat_tmp < 0) ? 0 : 0xff; \
    (u8)sat_tmp; \
    })

#define do_usat16(op0, op, op1) ({ \
    s32 sat_tmp; \
    sat_tmp = (s32)(u16)(op0) op (s32)(u16)(op1); \
    if (sat_tmp != (u16)sat_tmp) \
      sat_tmp = (sat_tmp < 0) ? 0 : 0xffff; \
    (u16)sat_tmp; \
    })

#define do_usat32(op0, op, op1) ({ \
    s64 sat_tmp; \
    sat_tmp = (s64)(u32)(op0) op (s64)(u32)(op1); \
    if (sat_tmp != (u32)sat_tmp) \
      sat_tmp = (sat_tmp < 0) ? 0 : 0xffffffff; \
    (u32)sat_tmp; \
    })

#define muls32_64(a, b) ((s64)(s32)a * (s64)(s32)b)

#define SIGN64 (1ull << 63)
static inline s64
double_sat64 (s64 a)
{
    if ((a ^ (a << 1)) & SIGN64) {
	return (a >> 63) ^ ~SIGN64;
    }
    return a << 1;
}

#define do_sat64(sat_op0, op, sat_op1) ({ \
    u64 op0 = sat_op0; \
    u64 op1 = sat_op1; \
    s64 sat_tmp = op0 op op1; \
    switch (1 op 2) { \
      case 3: /* Plus */ \
	if (((op0 ^ op1) & SIGN64) == 0 && ((op0 ^ sat_tmp) & SIGN64)) \
	    sat_tmp = ((s64)op0 >> 63) ^ ~SIGN64; \
	break; \
      case -1: /* Minus */ \
	if (((op0 ^ op1) & SIGN64) != 0 && ((op0 ^ sat_tmp) & SIGN64)) \
	    sat_tmp = ((s64)op0 >> 63) ^ ~SIGN64; \
	break; \
      default: \
	abort(); \
      } \
    sat_tmp; \
    })

#define do_usat64(sat_op0, op, sat_op1) ({ \
    u64 op0 = sat_op0; \
    u64 op1 = sat_op1; \
    u64 sat_tmp = op0 op op1; \
    switch (1 op 2) { \
      case 3: /* Plus */ \
	if (sat_tmp < op0) \
	    sat_tmp = (u64)-1; \
	break; \
      case -1: /* Minus */ \
	if (op0 < op1) \
	    sat_tmp = 0; \
	break; \
      default: \
	abort(); \
      } \
    sat_tmp; \
    })
